@@ -1,4 +1,4 @@
-/* $Id: rbldnsd.c,v 1.106 2005/04/16 14:02:04 mjt Exp $
+/* $Id: rbldnsd.c,v 1.117 2006-11-24 16:35:28 mjt Exp $
  * rbldnsd: main program
  */
 
@@ -21,33 +21,34 @@
 #include <time.h>
 #include <sys/time.h>	/* some systems can't include time.h and sys/time.h */
 #include <fcntl.h>
-#include <setjmp.h>
 #include <sys/wait.h>
 #include "rbldnsd.h"
-#include "rbldnsd_hooks.h"
 
-#ifndef NOSELECT_H
+#ifndef NO_SELECT_H
 # include <sys/select.h>
 #endif
-#ifndef NOPOLL
+#ifndef NO_POLL
 # include <sys/poll.h>
 #endif
-#ifndef NOMEMINFO
+#ifndef NO_MEMINFO
 # include <malloc.h>
 #endif
-#ifndef NOTIMES
+#ifndef NO_TIMES
 # include <sys/times.h>
 #endif
-#ifndef NOSTDINT_H
+#ifndef NO_STDINT_H
 /* if system have stdint.h, assume it have inttypes.h too */
 # include <inttypes.h>
 #endif
-#ifndef NOSTATS
-# ifndef NOIOVEC
+#ifndef NO_STATS
+# ifndef NO_IOVEC
 #  include <sys/uio.h>
 #  define STATS_IPC_IOVEC 1
 # endif
 #endif
+#ifndef NO_DSO
+# include <dlfcn.h>
+#endif
 
 #ifndef NI_MAXHOST
 # define NI_MAXHOST 1025
@@ -87,11 +88,12 @@ void error(int errnum, const char *fmt, ...) {
 static unsigned recheck = 60;	/* interval between checks for reload */
 static int initialized;		/* 1 when initialized */
 static char *logfile;		/* log file name */
-#ifndef NOSTATS
+#ifndef NO_STATS
 static char *statsfile;		/* statistics file */
 static int stats_relative;	/* dump relative, not absolute, stats */
 #endif
 int accept_in_cidr;		/* accept 127.0.0.1/8-"style" CIDRs */
+int nouncompress;		/* disable on-the-fly decompression */
 unsigned def_ttl = 35*60;	/* default record TTL 35m */
 unsigned min_ttl, max_ttl;	/* TTL constraints */
 const char def_rr[5] = "\177\0\0\2\0";		/* default A RR */
@@ -106,11 +108,13 @@ static int numzones;		/* number of zones in zonelist */
 int lazy;			/* don't return AUTH section by default */
 static int fork_on_reload;
   /* >0 - perform fork on reloads, <0 - this is a child of reloading parent */
-static pid_t bgq_pid;		/* pid of bg query process */
-static int ipc_fd = -1;		/* pipe FD for IPC */
 #if STATS_IPC_IOVEC
 static struct iovec *stats_iov;
 #endif
+#ifndef NO_DSO
+int (*hook_reload_check)(), (*hook_reload)();
+int (*hook_query_access)(), (*hook_query_result)();
+#endif
 
 /* a list of zonetypes. */
 const struct dstype *ds_types[] = {
@@ -119,7 +123,7 @@ const struct dstype *ds_types[] = {
   dstype(ip4trie),
   dstype(dnset),
 #ifdef DNHASH
-  dstype(dnhasn),
+  dstype(dnhash),
 #endif
   dstype(combined),
   dstype(generic),
@@ -127,6 +131,8 @@ const struct dstype *ds_types[] = {
   NULL
 };
 
+static int do_reload(int do_fork);
+
 static int satoi(const char *s) {
   int n = 0;
   if (*s < '0' || *s > '9') return -1;
@@ -135,65 +141,17 @@ static int satoi(const char *s) {
   return *s ? -1 : n;
 }
 
-static int do_reload(void) {
-  int r;
-  char ibuf[150];
-  int ip;
-#ifndef NOTIMES
-  struct tms tms;
-  clock_t utm, etm;
-#ifndef HZ
-  static clock_t HZ;
-  if (!HZ)
-    HZ = sysconf(_SC_CLK_TCK);
-#endif
-  etm = times(&tms);
-  utm = tms.tms_utime;
-#endif /* NOTIMES */
-
-  r = reloadzones(zonelist);
-  if (!r)
-    return 1;
-
-#ifdef do_hook_reload
-  hook_reload(zonelist);
-#endif
-
-  ip = ssprintf(ibuf, sizeof(ibuf), "zones reloaded");
-#ifndef NOTIMES
-  etm = times(&tms) - etm;
-  utm = tms.tms_utime - utm;
-# define sec(tm) (unsigned long)(tm/HZ), (unsigned long)((tm*100/HZ)%100)
-  ip += ssprintf(ibuf + ip, sizeof(ibuf) - ip,
-        ", time %lu.%lue/%lu.%luu sec", sec(etm), sec(utm));
-# undef sec
-#endif /* NOTIMES */
-#ifndef NOMEMINFO
-  {
-    struct mallinfo mi = mallinfo();
-# define kb(x) ((mi.x + 512)>>10)
-    ip += ssprintf(ibuf + ip, sizeof(ibuf) - ip,
-          ", mem arena=%d free=%d mmap=%d Kb",
-          kb(arena), kb(fordblks), kb(hblkhd));
-# undef kb
-  }
-#endif /* NOMEMINFO */
-  dslog(LOG_INFO, 0, ibuf);
-
-  return r < 0 ? 0 : 1;
-}
-
 static void NORETURN usage(int exitcode) {
    const struct dstype **dstp;
    printf(
-"%s: rbl dns daemon version %s%s\n"
+"%s: rbl dns daemon version %s\n"
 "Usage is: %s options zonespec...\n"
 "where options are:\n"
 " -u user[:group] - run as this user:group (rbldns)\n"
 " -r rootdir - chroot to this directory\n"
 " -w workdir - working directory with zone files\n"
 " -b address[/port] - bind to (listen on) this address (required)\n"
-#ifndef NOIPv6
+#ifndef NO_IPv6
 " -4 - use IPv4 socket type\n"
 " -6 - use IPv6 socket type\n"
 #endif
@@ -209,22 +167,28 @@ static void NORETURN usage(int exitcode) {
 "  during reload (may double memory requiriments)\n"
 " -q - quickstart, load zones after backgrounding\n"
 " -l [+]logfile - log queries and answers to this file (+ for unbuffered)\n"
-#ifndef NOSTATS
+#ifndef NO_STATS
 " -s [+]statsfile - write a line with short statistics summary into this\n"
 "  file every `check' (-c) secounds, for rrdtool-like applications\n"
 "  (+ to log relative, not absolute, statistics counters)\n"
 #endif
-" -a (experimental) - _omit_ AUTH section when constructing reply,\n"
-"  do not return list of auth nameservers in default replies, only\n"
-"  return NS info when explicitly asked\n"
-#ifdef do_hook_getopt
-" -H local_hook_options - process custom options (for custom builds)\n"
+" -a - omit AUTH section from regular replies, do not return list of\n"
+"  nameservers, but only return NS info when explicitly asked.\n"
+"  This is an equivalent of bind9 \"minimal-answers\" setting.\n"
+"  In future versions this mode will be the default.\n"
+" -A - put AUTH section in every reply.\n"
+#ifndef NO_ZLIB
+" -C - disable on-the-fly decompression of dataset files\n"
+#endif
+#ifndef NO_DZO
+" -x extension - load given extension module (.so file)\n"
+" -X extarg - pass extarg to extension init routine\n"
 #endif
 " -d - dump all zones in BIND format to standard output and exit\n"
 "each zone specified using `name:type:file,file...'\n"
 "syntax, repeated names constitute the same zone.\n"
 "Available dataset types:\n"
-, progname, version, hook_info, progname);
+, progname, version, progname);
   for(dstp = ds_types; *dstp; ++dstp)
     printf(" %s - %s\n", (*dstp)->dst_name, (*dstp)->dst_descr);
   exit(exitcode);
@@ -238,7 +202,7 @@ static volatile int signalled;
 #define SIGNALLED_ZSTATS	0x10
 #define SIGNALLED_TERM		0x20
 
-#ifdef NOIPv6
+#ifdef NO_IPv6
 static void newsocket(struct sockaddr_in *sin) {
   int fd;
   const char *host = ip4atos(ntohl(sin->sin_addr.s_addr));
@@ -284,7 +248,7 @@ initsockets(const char *bindaddr[MAXSOCK], int nba, int UNUSED family) {
   char *host, *serv;
   const char *ba;
 
-#ifdef NOIPv6
+#ifdef NO_IPv6
 
   struct sockaddr_in sin;
   ip4addr_t sinaddr;
@@ -322,7 +286,7 @@ initsockets(const char *bindaddr[MAXSOCK], int nba, int UNUSED family) {
         error(0, "missing host part in bind address `%.60s'", ba);
     }
 
-#ifdef NOIPv6
+#ifdef NO_IPv6
 
     if (!serv || !*serv)
       sin.sin_port = port;
@@ -374,7 +338,7 @@ initsockets(const char *bindaddr[MAXSOCK], int nba, int UNUSED family) {
   for (i = 0; i < numsock; ++i) {
     x = 65536;
     do
-      if (setsockopt(sock[i], SOL_SOCKET, SO_RCVBUF, &x, sizeof x) == 0)
+      if (setsockopt(sock[i], SOL_SOCKET, SO_RCVBUF, (void*)&x, sizeof x) == 0)
         break;
     while ((x -= (x >> 5)) >= 1024);
   }
@@ -391,6 +355,12 @@ static void init(int argc, char **argv) {
   gid_t gid = 0;
   int nodaemon = 0, quickstart = 0, dump = 0, nover = 0, forkon = 0;
   int family = AF_UNSPEC;
+  int cfd = -1;
+  const struct zone *z;
+#ifndef NO_DSO
+  char *ext = NULL, *extarg = NULL;
+  int (*extinit)(const char *arg, struct zone *zonelist) = NULL;
+#endif
 
   if ((progname = strrchr(argv[0], '/')) != NULL)
     argv[0] = ++progname;
@@ -399,7 +369,7 @@ static void init(int argc, char **argv) {
 
   if (argc <= 1) usage(1);
 
-  while((c = getopt(argc, argv, "u:r:b:w:t:c:p:nel:qs:h46dvafH:")) != EOF)
+  while((c = getopt(argc, argv, "u:r:b:w:t:c:p:nel:qs:h46dvaAfCx:X:")) != EOF)
     switch(c) {
     case 'u': user = optarg; break;
     case 'r': rootdir = optarg; break;
@@ -408,7 +378,7 @@ static void init(int argc, char **argv) {
         error(0, "too many addresses to listen on (%d max)", MAXSOCK);
       bindaddr[nba++] = optarg;
       break;
-#ifndef NOIPv6
+#ifndef NO_IPv6
     case '4': family = AF_INET; break;
     case '6': family = AF_INET6; break;
 #else
@@ -459,7 +429,7 @@ static void init(int argc, char **argv) {
       break;
 break;
     case 's':
-#ifdef NOSTATS
+#ifdef NO_STATS
       fprintf(stderr,
         "%s: warning: no statistics counters support is compiled in\n",
         progname);
@@ -479,14 +449,16 @@ break;
       break;
     case 'v': show_version = nover++ ? NULL : "rbldnsd"; break;
     case 'a': lazy = 1; break;
+    case 'A': lazy = 0; break;
     case 'f': forkon = 1; break;
-    case 'H':
-#ifdef do_hook_getopt
-      if (hook_getopt(optarg) != 0)
-        error(0, "error processing custom option `%s'", optarg);
-      break;
+    case 'C': nouncompress = 1; break;
+#ifndef NO_DSO
+    case 'x': ext = optarg; break;
+    case 'X': extarg = optarg; break;
 #else
-      error(0, "no custom option processing is compiled in");
+    case 'x':
+    case 'X':
+      error(0, "extension support is not compiled in");
 #endif
     case 'h': usage(0);
     default: error(0, "type `%.50s -h' for help", progname);
@@ -498,7 +470,6 @@ break;
 
 #ifndef NO_MASTER_DUMP
   if (dump) {
-    struct zone *z;
     time_t now;
     logto = LOGTO_STDERR;
     for(c = 0; c < argc; ++c)
@@ -508,7 +479,7 @@ break;
       error(errno, "unable to chroot to %.50s", rootdir);
     if (workdir && chdir(workdir) < 0)
       error(errno, "unable to chdir to %.50s", workdir);
-    if (!do_reload())
+    if (!do_reload(0))
       error(0, "zone loading errors, aborting");
     now = time(NULL);
     printf("; zone dump made %s", ctime(&now));
@@ -537,7 +508,7 @@ break;
       if (read(pfd[0], &c, 1) < 1) exit(1);
       else exit(0);
     }
-    ipc_fd = pfd[1];
+    cfd = pfd[1];
     close(pfd[0]);
     openlog(progname, LOG_PID|LOG_NDELAY, LOG_DAEMON);
     logto = LOGTO_STDERR|LOGTO_SYSLOG;
@@ -546,22 +517,35 @@ break;
 
   initsockets(bindaddr, nba, family);
 
+#ifndef NO_DSO
+  if (ext) {
+    void *handle = dlopen(ext, RTLD_NOW);
+    if (!handle)
+      error(0, "unable to load extension `%s': %s", ext, dlerror());
+    extinit = dlsym(handle, "rbldnsd_extension_init");
+    if (!extinit)
+      error(0, "unable to find extension init routine in `%s'", ext);
+  }
+#endif
+
   if (!user && !(uid = getuid()))
     user = "rbldns";
 
-  if (user && (p = strchr(user, ':')) != NULL)
-    *p++ = '\0';
   if (!user)
     p = NULL;
-  else if ((c = satoi(user)) >= 0)
-    uid = c, gid = c;
   else {
-    struct passwd *pw = getpwnam(user);
-    if (!pw)
-      error(0, "unknown user `%s'", user);
-    uid = pw->pw_uid;
-    gid = pw->pw_gid;
-    endpwent();
+    if ((p = strchr(user, ':')) != NULL)
+      *p++ = '\0';
+    if ((c = satoi(user)) >= 0)
+      uid = c, gid = c;
+    else {
+      struct passwd *pw = getpwnam(user);
+      if (!pw)
+        error(0, "unknown user `%s'", user);
+      uid = pw->pw_uid;
+      gid = pw->pw_gid;
+      endpwent();
+    }
   }
   if (!uid)
     error(0, "daemon should not run as root, specify -u option");
@@ -600,35 +584,34 @@ break;
   for(c = 0; c < argc; ++c)
     zonelist = addzone(zonelist, argv[c]);
   init_zones_caches(zonelist);
-#ifdef do_hook_init
-  if (hook_init(zonelist) != 0) error(0, "error processing init hook");
+
+#ifndef NO_DSO
+  if (extinit && extinit(extarg, zonelist) != 0)
+    error(0, "unable to iniitialize extension `%s'", ext);
 #endif
 
-  if (!quickstart && !do_reload())
+  if (!quickstart && !do_reload(0))
     error(0, "zone loading errors, aborting");
 
-  { const struct zone *z;
-    for(c = 0, z = zonelist; z; z = z->z_next)
-     ++c;
-    numzones = c;
-  }
+  /* count number of zones */
+  for(c = 0, z = zonelist; z; z = z->z_next)
+    ++c;
+  numzones = c;
+
 #if STATS_IPC_IOVEC
   stats_iov = (struct iovec *)emalloc(numzones * sizeof(struct iovec));
-  { struct zone *z;
-    for(c = 0, z = zonelist; z; z = z->z_next, ++c) {
-      stats_iov[c].iov_base = (char*)&z->z_stats;
-      stats_iov[c].iov_len = sizeof(z->z_stats);
-    }
+  for(c = 0, z = zonelist; z; z = z->z_next, ++c) {
+    stats_iov[c].iov_base = (char*)&z->z_stats;
+    stats_iov[c].iov_len = sizeof(z->z_stats);
   }
 #endif
   dslog(LOG_INFO, 0, "rbldnsd version %s started (%d socket(s), %d zone(s))",
         version, numsock, numzones);
   initialized = 1;
 
-  if (ipc_fd >= 0) {
-    write(ipc_fd, "", 1);
-    close(ipc_fd);
-    ipc_fd = -1;
+  if (cfd >= 0) {
+    write(cfd, "", 1);
+    close(cfd);
     close(0); close(2);
     if (!flog) close(1);
     setsid();
@@ -636,7 +619,7 @@ break;
   }
 
   if (quickstart)
-    do_reload();
+    do_reload(0);
 
   /* only set "main" fork_on_reload after first reload */
   fork_on_reload = forkon;
@@ -648,10 +631,12 @@ static void sighandler(int sig) {
     signalled |= SIGNALLED_RELOG|SIGNALLED_RELOAD;
     break;
   case SIGALRM:
+#ifndef HAVE_SETITIMER
     alarm(recheck);
+#endif
     signalled |= SIGNALLED_RELOAD|SIGNALLED_SSTATS;
     break;
-#ifndef NOSTATS
+#ifndef NO_STATS
   case SIGUSR1:
     signalled |= SIGNALLED_LSTATS|SIGNALLED_SSTATS;
     break;
@@ -679,7 +664,7 @@ static void setup_signals(void) {
   sigaddset(&ssblock, SIGHUP);
   sigaction(SIGALRM, &sa, NULL);
   sigaddset(&ssblock, SIGALRM);
-#ifndef NOSTATS
+#ifndef NO_STATS
   sigaction(SIGUSR1, &sa, NULL);
   sigaddset(&ssblock, SIGUSR1);
   sigaction(SIGUSR2, &sa, NULL);
@@ -690,7 +675,7 @@ static void setup_signals(void) {
   signal(SIGPIPE, SIG_IGN);	/* in case logfile is FIFO */
 }
 
-#ifndef NOSTATS
+#ifndef NO_STATS
 
 struct dnsstats gstats;
 static struct dnsstats gptot;
@@ -829,77 +814,197 @@ static void reopenlog(void) {
   }
 }
 
-/* two-process reload:
- * we call reloadzones(), which first check whenever
- * any files changed, and if yes, it calls start_loading()
- * (for every dataset with changed files) and performs all
- * necessary updates.
- * In first call to start_loading(), we fork the child.
- * In child, we return (with longjmp) back into do_signalled(),
- * setjmp() returning !=0, and we just continue servicing requests
- * at this point with fork_at_reload set to -1 and with all
- * fancy signals (SIGALRM, SIGUSR?, SIGHUP) ignored.
- * In parent, real reload continues, do_reload completes,
- * and we check for bgq_pid: this is our child, we kill it,
- * and read all the counters from the pipe we opened.
- * In child (note fork_on_reload is <0), upon receiving SIGTERM,
- * we write all the stats into pipe and terminate silently.
- * All signals are still blocked during the whole reload in parent,
- * we unblock them only after we successefully reaped the child.
- * The only possible problem is when the parent gets killed (either
- * SIGKILL or crash) while child is running - in this case child will
- * stay running forever.  It may be a good idea to setup SIGALRM
- * handler in child to check for parent periodically...
- */
-
-static jmp_buf reload_ctx; /* longjmp in start_loading() */
-
-int start_loading() {
-  pid_t cpid;
-  int pfd[2];
-  if (!fork_on_reload || bgq_pid) return 0;
-  if (flog && !flushlog)
-    fflush(flog);
-  if (pipe(pfd) < 0) {
-    bgq_pid = -1;
-    return 0;
+static void check_expires(void) {
+  struct zone *zone;
+  time_t now = time(NULL);
+  for (zone = zonelist; zone; zone = zone->z_next) {
+    if (!zone->z_stamp)
+      continue;
+    if (zone->z_expires && zone->z_expires < now) {
+      zlog(LOG_WARNING, zone, "zone data expired, zone will not be serviced");
+      zone->z_stamp = 0;
+    }
   }
-  cpid = fork();
-  if (cpid < 0) {
-    close(pfd[0]);
-    close(pfd[1]);
-    bgq_pid = -1;
-    return 0;
-  }
-  if (!cpid) { /* child, continue answering queries */
-    fork_on_reload = -1;
-    signal(SIGALRM, SIG_IGN);
-    signal(SIGHUP, SIG_IGN);
-#ifndef NOSTATS
-    signal(SIGUSR1, SIG_IGN);
-    signal(SIGUSR2, SIG_IGN);
+}
+
+static int do_reload(int do_fork) {
+  int r;
+  char ibuf[150];
+  int ip;
+  struct dataset *ds;
+  struct zone *zone;
+  pid_t cpid = 0;	/* child pid; =0 to make gcc happy */
+  int cfd = 0;		/* child stats fd; =0 to make gcc happy */
+#ifndef NO_TIMES
+  struct tms tms;
+  clock_t utm, etm;
+#ifndef HZ
+  static clock_t HZ;
 #endif
-    close(pfd[0]);
-    ipc_fd = pfd[1];
-    longjmp(reload_ctx, 1);
+#endif /* NO_TIMES */
+
+  ds = nextdataset2reload(NULL);
+  if (!ds && call_hook(reload_check, (zonelist)) == 0) {
+    check_expires();
+    return 1;	/* nothing to reload */
   }
-  close(pfd[1]);
-  ipc_fd = pfd[0];
-  bgq_pid = cpid;
-  return 0;
+
+  if (do_fork) {
+    int pfd[2];
+    if (flog && !flushlog)
+      fflush(flog);
+    /* forking reload. if anything fails, just do a non-forking one */
+    if (pipe(pfd) < 0)
+      do_fork = 0;
+    else if ((cpid = fork()) < 0) {	/* fork failed, close the pipe */
+      close(pfd[0]);
+      close(pfd[1]);
+      do_fork = 0;
+    }
+    else if (!cpid) {	/* child, continue answering queries */
+      signal(SIGALRM, SIG_IGN);
+      signal(SIGHUP, SIG_IGN);
+#ifndef NO_STATS
+      signal(SIGUSR1, SIG_IGN);
+      signal(SIGUSR2, SIG_IGN);
+#endif
+      close(pfd[0]);
+      /* set up the fd#1 to write stats later on SIGTERM */
+      if (pfd[1] != 1) {
+        dup2(pfd[1], 1);
+        close(pfd[1]);
+      }
+      fork_on_reload = -1;
+      return 1;
+    }
+    else {
+      close(pfd[1]);
+      cfd = pfd[0];
+    }
+  }
+
+#ifndef NO_TIMES
+#ifndef HZ
+  if (!HZ)
+    HZ = sysconf(_SC_CLK_TCK);
+#endif
+  etm = times(&tms);
+  utm = tms.tms_utime;
+#endif /* NO_TIMES */
+
+  r = 1;
+  while(ds) {
+    if (!loaddataset(ds))
+      r = 0;
+    ds = nextdataset2reload(ds);
+  }
+
+  for (zone = zonelist; zone; zone = zone->z_next) {
+    time_t stamp = 0;
+    time_t expires = 0;
+    const struct dssoa *dssoa = NULL;
+    const struct dsns *dsns = NULL;
+    unsigned nsttl = 0;
+    struct dslist *dsl;
+
+    for(dsl = zone->z_dsl; dsl; dsl = dsl->dsl_next) {
+      const struct dataset *ds = dsl->dsl_ds;
+      if (!ds->ds_stamp) {
+        stamp = 0;
+        break;
+      }
+      if (stamp < ds->ds_stamp)
+        stamp = ds->ds_stamp;
+      if (ds->ds_expires && (!expires || expires > ds->ds_expires))
+        expires = ds->ds_expires;
+      if (!dssoa)
+        dssoa = ds->ds_dssoa;
+      if (!dsns)
+        dsns = ds->ds_dsns, nsttl = ds->ds_nsttl;
+    }
+
+    zone->z_expires = expires;
+    zone->z_stamp = stamp;
+    if (!stamp) {
+      zlog(LOG_WARNING, zone,
+           "not all datasets are loaded, zone will not be serviced");
+      r = 0;
+    }
+    else if (!update_zone_soa(zone, dssoa) ||
+             !update_zone_ns(zone, dsns, nsttl, zonelist))
+      zlog(LOG_WARNING, zone,
+           "NS or SOA RRs are too long, will be ignored");
+  }
+
+  if (call_hook(reload, (zonelist)) != 0)
+    r = 0;
+
+  ip = ssprintf(ibuf, sizeof(ibuf), "zones reloaded");
+#ifndef NO_TIMES
+  etm = times(&tms) - etm;
+  utm = tms.tms_utime - utm;
+# define sec(tm) (unsigned long)(tm/HZ), (unsigned long)((tm*100/HZ)%100)
+  ip += ssprintf(ibuf + ip, sizeof(ibuf) - ip,
+        ", time %lu.%lue/%lu.%luu sec", sec(etm), sec(utm));
+# undef sec
+#endif /* NO_TIMES */
+#ifndef NO_MEMINFO
+  {
+    struct mallinfo mi = mallinfo();
+# define kb(x) ((mi.x + 512)>>10)
+    ip += ssprintf(ibuf + ip, sizeof(ibuf) - ip,
+          ", mem arena=%d free=%d mmap=%d Kb",
+          kb(arena), kb(fordblks), kb(hblkhd));
+# undef kb
+  }
+#endif /* NO_MEMINFO */
+  dslog(LOG_INFO, 0, ibuf);
+
+  check_expires();
+
+  /* ok, (something) loaded. */
+
+  if (do_fork) {
+    /* here we should notify query-answering child (send SIGTERM to it),
+     * and wait for it to complete.
+     * Unfortunately at least on linux, the SIGTERM sometimes gets ignored
+     * by the child process, so we're trying several times here, in a loop.
+     */
+    int s, n;
+    fd_set fds;
+    struct timeval tv;
+
+    for(n = 1; ++n;) {
+      if (kill(cpid, SIGTERM) != 0)
+        dslog(LOG_WARNING, 0, "kill(qchild): %s", strerror(errno));
+      FD_ZERO(&fds);
+      FD_SET(cfd, &fds);
+      tv.tv_sec = 0;
+      tv.tv_usec = 500000;
+      s = select(cfd+1, &fds, NULL, NULL, &tv);
+      if (s > 0) break;
+      dslog(LOG_WARNING, 0, "waiting for qchild process: %s, retrying",
+            s ? strerror(errno) : "timeout");
+    }
+    ipc_read_stats(cfd);
+    close(cfd);
+    wait(&s);
+  }
+
+  return r;
 }
 
 static void do_signalled(void) {
   sigprocmask(SIG_SETMASK, &ssblock, NULL);
   if (signalled & SIGNALLED_TERM) {
     if (fork_on_reload < 0) { /* this is a temp child; dump stats and exit */
-      ipc_write_stats(ipc_fd);
+      ipc_write_stats(1);
       if (flog && !flushlog)
         fflush(flog);
       _exit(0);
     }
     dslog(LOG_INFO, 0, "terminating");
-#ifndef NOSTATS
+#ifndef NO_STATS
     if (statsfile)
       dumpstats();
     logstats(0);
@@ -908,7 +1013,7 @@ static void do_signalled(void) {
 #endif
     exit(0);
   }
-#ifndef NOSTATS
+#ifndef NO_STATS
   if (signalled & SIGNALLED_SSTATS && statsfile)
     dumpstats();
   if (signalled & SIGNALLED_LSTATS) {
@@ -919,42 +1024,13 @@ static void do_signalled(void) {
 #endif
   if (signalled & SIGNALLED_RELOG)
     reopenlog();
-  if (signalled & SIGNALLED_RELOAD) {
-    if (!fork_on_reload) /* normal reload */
-      do_reload();
-    else /* else two-process reload */
-    if (!setjmp(reload_ctx)) {
-      do_reload();
-      if (bgq_pid > 0) {
-        int s, n;
-        fd_set fds;
-        struct timeval tv;
-
-        for(n = 1; ++n;) {
-          if (kill(bgq_pid, SIGTERM) != 0)
-            dslog(LOG_WARNING, 0, "kill(qchild): %s", strerror(errno));
-          FD_ZERO(&fds);
-          FD_SET(ipc_fd, &fds);
-          tv.tv_sec = 0;
-          tv.tv_usec = 500000;
-          s = select(ipc_fd+1, &fds, NULL, NULL, &tv);
-          if (s > 0) break;
-          dslog(LOG_WARNING, 0, "waiting for qchild process: %s, retrying",
-                s ? strerror(errno) : "timeout");
-        }
-        ipc_read_stats(ipc_fd);
-        close(ipc_fd);
-        ipc_fd = -1;
-        wait(&s);
-      } /* bgq_pid > 0 */
-      bgq_pid = 0;
-    } /* 2process reload, parent */
-  }
+  if (signalled & SIGNALLED_RELOAD)
+    do_reload(fork_on_reload);
   signalled = 0;
   sigprocmask(SIG_SETMASK, &ssempty, NULL);
 }
 
-#ifndef NOIPv6
+#ifndef NO_IPv6
 static struct sockaddr_storage peer_sa;
 #else
 static struct sockaddr_in peer_sa;
@@ -965,7 +1041,7 @@ static void request(int fd) {
   int q, r;
   socklen_t salen = sizeof(peer_sa);
 
-  q = recvfrom(fd, pkt.p_buf, sizeof(pkt.p_buf), 0,
+  q = recvfrom(fd, (void*)pkt.p_buf, sizeof(pkt.p_buf), 0,
                (struct sockaddr *)&peer_sa, &salen);
   if (q <= 0)			/* interrupted? */
     return;
@@ -978,7 +1054,8 @@ static void request(int fd) {
     logreply(&pkt, flog, flushlog);
 
   /* finally, send a reply */
-  while(sendto(fd, pkt.p_buf, r, 0, (struct sockaddr *)&peer_sa, salen) < 0)
+  while(sendto(fd, (void*)pkt.p_buf, r, 0,
+               (struct sockaddr *)&peer_sa, salen) < 0)
     if (errno != EINTR) break;
 
 }
@@ -987,8 +1064,18 @@ int main(int argc, char **argv) {
   init(argc, argv);
   setup_signals();
   reopenlog();
+#ifdef HAVE_SETITIMER
+  if (recheck) {
+    struct itimerval itv;
+    itv.it_interval.tv_sec  = itv.it_value.tv_sec  = recheck;
+    itv.it_interval.tv_usec = itv.it_value.tv_usec = 0;
+    if (setitimer(ITIMER_REAL, &itv, NULL) < 0)
+      error(errno, "unable to setitimer()");
+  }
+#else
   alarm(recheck);
-#ifndef NOSTATS
+#endif
+#ifndef NO_STATS
   stats_time = time(NULL);
   if (statsfile)
     dumpstats_z();
@@ -1006,7 +1093,7 @@ int main(int argc, char **argv) {
   }
   else {
     /* several sockets, do select/poll loop */
-#ifdef NOPOLL
+#ifdef NO_POLL
     fd_set rfds;
     int maxfd = 0;
     int *fdi, *fde = sock + numsock;
@@ -1026,7 +1113,7 @@ int main(int argc, char **argv) {
           request(*fdi);
       }
     }
-#else /* !NOPOLL */
+#else /* !NO_POLL */
     struct pollfd pfda[MAXSOCK];
     struct pollfd *pfdi, *pfde = pfda + numsock;
     int r;
@@ -1044,7 +1131,7 @@ int main(int argc, char **argv) {
         if (!--r) break;
       }
     }
-#endif /* NOPOLL */
+#endif /* NO_POLL */
   }
 }
 
