@@ -1,4 +1,4 @@
-/* $Id: rbldnsd.h,v 1.90 2005/04/27 12:27:17 mjt Exp $
+/* $Id: rbldnsd.h,v 1.97 2006/02/01 10:03:25 mjt Exp $
  * common rbldnsd #include header
  */
 
@@ -82,6 +82,7 @@ char *parse_uint32_nb(char *s, unsigned char nb[4]);
 char *parse_time(char *s, unsigned *tp);
 char *parse_time_nb(char *s, unsigned char nb[4]);
 char *parse_ttl(char *s, unsigned *ttlp, unsigned defttl);
+char *parse_timestamp(char *s, time_t *tsp);
 char *parse_dn(char *s, unsigned char *dn, unsigned *dnlenp);
 /* parse line in form :ip:text into rr
  * where first 4 bytes is ip in network byte order.
@@ -201,6 +202,7 @@ struct dataset {	/* ds */
   const struct dstype *ds_type;	/* type of this data */
   struct dsdata *ds_dsd;		/* type-specific data */
   time_t ds_stamp;			/* timestamp */
+  time_t ds_expires;			/* when the dataset expires if any */
   const char *ds_spec;			/* original specification */
   struct dsfile *ds_dsf;		/* list of files for this data */
   struct dssoa *ds_dssoa;		/* SOA record */
@@ -226,8 +228,8 @@ struct dslist {	/* dsl */
 struct zonesoa;
 struct zonens;
 
-#ifndef NOSTATS
-#if !defined(NOSTDINT_H)
+#ifndef NO_STATS
+#if !defined(NO_STDINT_H)
 typedef uint64_t dnscnt_t;
 #define PRI_DNSCNT PRIu64
 #elif SIZEOF_LONG < 8 && defined(SIZEOF_LONG_LONG)
@@ -244,10 +246,11 @@ struct dnsstats {
 extern struct dnsstats gstats;	/* global statistics counters */
 #endif /* NO_STATS */
 
-#define MAX_NS 20
+#define MAX_NS 32
 
 struct zone {	/* zone, list of zones */
   unsigned z_stamp;			/* timestamp, 0 if not loaded */
+  time_t z_expires;			/* when the zone expires if any */
   unsigned char z_dn[DNS_MAXDN+1];	/* zone domain name */
   unsigned z_dnlen;			/* length of z_dn */
   unsigned z_dnlab;			/* number of dn labels */
@@ -264,7 +267,7 @@ struct zone {	/* zone, list of zones */
   unsigned z_cns;			/* current NS in rotation */
   unsigned z_nglue;			/* number of glue records */
   struct zonens *z_zns;			/* pre-packed NS records */
-#ifndef NOSTATS
+#ifndef NO_STATS
   struct dnsstats z_stats;		/* statistic counters */
   struct dnsstats z_pstats;		/* for stats monitoring: prev values */
 #endif
@@ -320,7 +323,8 @@ void connectdataset(struct zone *zone,
 struct zone *newzone(struct zone **zonelist,
                      unsigned char *dn, unsigned dnlen,
                      struct mempool *mp);
-int reloadzones(struct zone *zonelist);
+struct dataset *nextdataset2reload(struct dataset *ds);
+int loaddataset(struct dataset *ds);
 
 struct dsctx {
   struct dataset *dsc_ds;	/* currently loading dataset */
@@ -337,16 +341,13 @@ void PRINTFLIKE(2,3) dsloaded(struct dsctx *dsc, const char *fmt, ...);
 void PRINTFLIKE(3,4)
 zlog(int level, const struct zone *zone, const char *fmt, ...);
 
-int readdslines(FILE *f, struct dataset *ds, struct dsctx *dsc);
-/* parse $SPECIAL */
-int ds_special(struct dataset *ds, char *line, struct dsctx *dsc);
-
 /* from rbldnsd_combined.c, special routine used inside ds_special() */
 int ds_combined_newset(struct dataset *ds, char *line, struct dsctx *dsc);
 
 extern unsigned def_ttl, min_ttl, max_ttl;
 extern const char def_rr[5];
 extern int accept_in_cidr;
+extern int nouncompress;
 extern struct dataset *g_dsacl;	/* global acl */
 
 extern const char *show_version; /* version.bind CH TXT */
@@ -366,10 +367,17 @@ int vssprintf(char *buf, int bufsz, const char *fmt, va_list ap);
 int PRINTFLIKE(3, 4) ssprintf(char *buf, int bufsz, const char *fmt, ...);
 
 /* a helper to shrink an array */
+/* some malloc implementations may return NULL on realloc()
+ * even if newsize is smaller than current size, so deal with
+ * this here.  Note that if realloc() fails, we're under memory
+ * pressure here, so next reload will probably hit ENOMEM anyway...
+ */
 #define SHRINK_ARRAY(type, arr, needed, allocated)	\
   if ((allocated) > (needed)) {				\
-     (arr) = trealloc(type, (arr), (needed));		\
-     (allocated) = (needed);				\
+     type *_temp = trealloc(type, (arr), (needed));	\
+     if (_temp) {					\
+       (arr) = _temp; (allocated) = (needed);		\
+     }							\
   }
 
 /* a helper macro to remove dups from a sorted array */
