@@ -1,4 +1,4 @@
-/* $Id: rbldnsd_util.c,v 1.47 2005/04/15 19:37:22 mjt Exp $
+/* $Id: rbldnsd_util.c,v 1.56 2007-11-16 17:52:15 mjt Exp $
  * Common utility routines for rbldnsd.
  */
 
@@ -83,6 +83,70 @@ char *parse_ttl(char *s, unsigned *ttlp, unsigned defttl) {
   return s;
 }
 
+static const unsigned char mday[12] = {
+  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
+};
+
+#define isleap(year) \
+  ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))
+
+static char *
+parse_tsp(char *s, unsigned *np, unsigned min, unsigned max, unsigned w) {
+  unsigned n = 0;
+  if (!digit(*s)) return NULL;
+  do n = n * 10 + d2n(*s++);
+  while(digit(*s) && --w);
+  if (n < min || n > max) return NULL;
+  if (*s == ':' || *s == '-') ++s;
+  *np = n;
+  return s;
+}
+
+char *parse_timestamp(char *s, time_t *tsp) {
+  unsigned year, mon, day, hour, min, sec;
+
+  if ((*s == '0' || *s == '-' || *s == ':') &&
+      (ISSPACE(s[1]) || !s[1])) {
+    *tsp = 0;
+    ++s;
+    SKIPSPACE(s);
+    return s;
+  }
+  if (!(s = parse_tsp(s, &year, 1970, 2038, 4))) return NULL;
+  if (!(s = parse_tsp(s, &mon, 1, 12, 2))) return NULL;
+  mon -= 1;
+  day = mon == 1 && isleap(year) ? 29 : mday[mon];
+  if (!(s = parse_tsp(s, &day, 1, day, 2))) return NULL;
+  hour = min = sec = 0;
+  if (*s && !ISSPACE(*s))
+    if (!(s = parse_tsp(s, &hour, 0, 23, 2))) return NULL;
+  if (*s && !ISSPACE(*s))
+    if (!(s = parse_tsp(s, &min, 0, 59, 2))) return NULL;
+  if (*s && !ISSPACE(*s))
+    if (!(s = parse_tsp(s, &sec, 0, 59, 2))) return NULL;
+  if (*s) {
+    if (!ISSPACE(*s)) return NULL;
+    ++s; SKIPSPACE(s);
+  }
+
+  {
+    unsigned y4 = (year / 4) - !(year & 3);
+    unsigned y100 = y4 / 25;
+    unsigned y400 = y100 / 4;
+    day =
+      365 * (year - 1970) +
+      (y4 - 492) - (y100 - 19) + (y400 - 4) +
+      day - 1;
+    if (isleap(year) && mon > 1)
+      ++day;
+    while(mon)
+      day += mday[--mon];
+    *tsp = ((day * 24 + hour) * 60 + min) * 60 + sec;
+  }
+
+  return s;
+}
+
 char *parse_dn(char *s, unsigned char *dn, unsigned *dnlenp) {
   char *n = s;
   unsigned l;
@@ -96,59 +160,6 @@ char *parse_dn(char *s, unsigned char *dn, unsigned *dnlenp) {
   return n;
 }
 
-int readdslines(FILE *f, struct dataset *ds, struct dsctx *dsc) {
-#define bufsiz 8192
-  char _buf[bufsiz+4], *line, *eol;
-#define buf (_buf+4)  /* keep room for 4 IP octets in addrtxt() */
-  int noeol = 0;
-  struct dataset *dscur = ds;
-  ds_linefn_t *linefn = dscur->ds_type->dst_linefn;
-
-  while(fgets(buf, bufsiz, f)) {
-    eol = buf + strlen(buf) - 1;
-    if (eol < buf) /* can this happen? */
-      continue;
-    if (noeol) { /* read parts of long line up to \n */
-      if (*eol == '\n')
-        noeol = 0;
-      continue;
-    }
-    ++dsc->dsc_lineno;
-    if (*eol == '\n')
-      --eol;
-    else if (feof(f)) {
-      dslog(LOG_WARNING, dsc, "incomplete last line (ignored)");
-      break;
-    }
-    else {
-      dswarn(dsc, "long line (truncated)");
-      noeol = 1; /* mark it to be read above */
-    }
-    /* skip whitespace */
-    line = buf;
-    SKIPSPACE(line);
-    while(eol >= line && ISSPACE(*eol))
-      --eol;
-    eol[1] = '\0';
-    if (line[0] == '$' ||
-        ((ISCOMMENT(line[0]) || line[0] == ':') && line[1] == '$')) {
-      int r = ds_special(ds, line[0] == '$' ? line + 1 : line + 2, dsc);
-      if (!r)
-        dswarn(dsc, "invalid or unrecognized special entry");
-      else if (r < 0)
-        return 0;
-      dscur = dsc->dsc_subset ? dsc->dsc_subset : ds;
-      linefn = dscur->ds_type->dst_linefn;
-      continue;
-    }
-    if (line[0] && !ISCOMMENT(line[0]))
-      if (!linefn(dscur, line, dsc))
-        return 0;
-  }
-  return 1;
-#undef buf
-}
-
 int parse_a_txt(char *str, const char **rrp, const char *def_rr,
                 struct dsctx *dsc) {
   char *rr;
@@ -166,7 +177,7 @@ int parse_a_txt(char *str, const char **rrp, const char *def_rr,
     if (*str == ':') {	/* A+TXT */
       ++str;
       SKIPSPACE(str);
-      rr = (unsigned char*)str - 4;
+      rr = str - 4;
       PACK32(rr, a);
     }
     else if (*str) {
@@ -183,7 +194,7 @@ int parse_a_txt(char *str, const char **rrp, const char *def_rr,
     }
   }
   else {
-    rr = (unsigned char*)str - 4;
+    rr = str - 4;
     memcpy(rr, def_rr, 4);
   }
   if (*str) {
@@ -208,6 +219,74 @@ unsigned unpack32(const unsigned char p[4]) {
   return n;
 }
 
+/* return pointer to the next word in line if first word is the same
+ * as word_lc (ignoring case), or NULL if not.
+ */
+char *firstword_lc(char *line, const char *word_lc) {
+  while(*word_lc)
+    if (dns_dnlc(*line) != *word_lc)
+      return NULL;
+    else
+      ++word_lc, ++line;
+  if (!ISSPACE(*line))
+    return NULL;
+  SKIPSPACE(line);
+ return line;
+}
+
+/* implement TXT substitutions.
+ * `sb' is a buffer where the result will be stored -
+ * at least 255 + 3 characters long */
+int txtsubst(char sb[TXTBUFSIZ], const char *txt,
+	     const char *s0, const struct dataset *ds) {
+  char *const *sn = ds->ds_subst;
+  unsigned sl;
+  char *const e = sb + 254;
+  char *lp = sb;
+  const char *s, *si, *sx;
+  if (txt[0] == '=')
+    sx = ++txt;
+  else if (sn[SUBST_BASE_TEMPLATE] && *sn[SUBST_BASE_TEMPLATE]) {
+    if (*txt) s0 = txt;
+    sx = s0;
+    txt = sn[SUBST_BASE_TEMPLATE];
+  }
+  else
+    sx = txt;
+  while(lp < e) {
+    if ((s = strchr(txt, '$')) == NULL)
+      s = (char*)txt + strlen(txt);
+    sl = s - txt;
+    if (lp + sl > e)
+      sl = e - lp;
+    memcpy(lp, txt, sl);
+    lp += sl;
+    if (!*s++) break;
+    if (*s == '$') { si = s++; sl = 1; }
+    else if (*s >= '0' && *s <= '9') { /* $n var */
+      si = sn[*s - '0'];
+      if (!si) { si = s - 1; sl = 2; }
+      else sl = strlen(si);
+      ++s;
+    }
+    else if (*s == '=') {
+      si = sx;
+      sl = strlen(si);
+      ++s;
+    }
+    else
+      sl = strlen(si = s0);
+    if (lp + sl > e) /* silently truncate TXT RR >255 bytes */
+      sl = e - lp;
+    memcpy(lp, si, sl);
+    lp += sl;
+    txt = s;
+  }
+  sl = lp - sb;
+  if (sl > 254) sl = 254;
+  return sl;
+}
+
 #ifndef NO_MASTER_DUMP
 
 void dump_ip4(ip4addr_t a, const char *rr, const struct dataset *ds, FILE *f) {
@@ -269,62 +348,21 @@ void dump_ip4range(ip4addr_t a, ip4addr_t b, const char *rr,
 
 }
 
-#endif
-
-/* implement TXT substitutions.
- * `sb' is a buffer where the result will be stored -
- * at least 255 + 3 characters long */
-int txtsubst(char sb[TXTBUFSIZ], const char *txt,
-	     const char *s0, const struct dataset *ds) {
-  char *const *sn = ds->ds_subst;
-  unsigned sl;
-  char *const e = sb + 254;
-  char *lp = sb;
-  const char *s, *si;
-  if (!s0) s0 = "$";
-  while(lp < e) {
-    if ((s = strchr(txt, '$')) == NULL)
-      s = (char*)txt + strlen(txt);
-    sl = s - txt;
-    if (lp + sl > e)
-      sl = e - lp;
-    memcpy(lp, txt, sl);
-    lp += sl;
-    if (!*s++) break;
-    if (*s == '$') { si = s++; sl = 1; }
-    else if (*s >= '0' && *s <= '9') { /* $1 var */
-      si = sn[*s - '0'];
-      if (!si) { si = s - 1; sl = 2; }
-      else sl = strlen(si);
-      ++s;
-    }
-    else
-      sl = strlen(si = s0);
-    if (lp + sl > e) /* silently truncate TXT RR >255 bytes */
-      sl = e - lp;
-    memcpy(lp, si, sl);
-    lp += sl;
-    txt = s;
-  }
-  sl = lp - sb;
-  if (sl > 254) sl = 254;
-  return sl;
-}
-
 void
-dump_a_txt(const char *name, const unsigned char *rr,
+dump_a_txt(const char *name, const char *rr,
            const char *subst, const struct dataset *ds, FILE *f) {
   if (!rr)
     fprintf(f, "%s\tCNAME\texcluded\n", name);
   else {
-    fprintf(f, "%s\tA\t%u.%u.%u.%u\n",
-            name, rr[0], rr[1], rr[2], rr[3]);
-    if (rr[4]) {
-      char txt[TXTBUFSIZ];
+    const unsigned char *a = (const unsigned char*)rr;
+    char sb[TXTBUFSIZ];
+    unsigned sl = txtsubst(sb, rr + 4, subst, ds);
+    fprintf(f, "%s\tA\t%u.%u.%u.%u\n", name, a[0], a[1], a[2], a[3]);
+    if (sl) {
       char *p, *n;
-      txt[txtsubst(txt, rr + 4, subst, ds)] = '\0';
+      sb[sl] = '\0';
       fprintf(f, "\tTXT\t\"");
-      for(p = txt; (n = strchr(p, '"')) != NULL; p = n + 1) {
+      for(p = sb; (n = strchr(p, '"')) != NULL; p = n + 1) {
         fwrite(p, 1, n - p, f);
         putc('\\', f); putc('"', f);
       }
@@ -333,6 +371,8 @@ dump_a_txt(const char *name, const unsigned char *rr,
   }
 }
 
+#endif
+
 char *emalloc(unsigned size) {
   void *ptr = malloc(size);
   if (!ptr)
@@ -377,7 +417,7 @@ char *estrdup(const char *str) {
 
 int vssprintf(char *buf, int bufsz, const char *fmt, va_list ap) {
   int r = vsnprintf(buf, bufsz, fmt, ap);
-  return r < 0 ? 0 : r >= bufsz ? buf[bufsiz-1] = '\0', bufsz - 1 : r;
+  return r < 0 ? 0 : r >= bufsz ? buf[bufsz-1] = '\0', bufsz - 1 : r;
 }
 
 int ssprintf(char *buf, int bufsz, const char *fmt, ...) {
@@ -434,20 +474,20 @@ vdslog(int level, struct dsctx *dsc, const char *fmt, va_list ap) {
     write(1, buf, l);
 }
 
-void dslog(int level, struct dsctx *dsl, const char *fmt, ...) {
+void dslog(int level, struct dsctx *dsc, const char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
-  vdslog(level, dsl, fmt, ap);
+  vdslog(level, dsc, fmt, ap);
   va_end(ap);
 }
 
 #define MAXWARN 5
 
-void dswarn(struct dsctx *dsl, const char *fmt, ...) {
-  if (++dsl->dsc_warns <= MAXWARN) { /* prevent syslog flood */
+void dswarn(struct dsctx *dsc, const char *fmt, ...) {
+  if (++dsc->dsc_warns <= MAXWARN) { /* prevent syslog flood */
     va_list ap;
     va_start(ap, fmt);
-    vdslog(LOG_WARNING, dsl, fmt, ap);
+    vdslog(LOG_WARNING, dsc, fmt, ap);
     va_end(ap);
   }
 }
