@@ -1,4 +1,4 @@
-/* $Id: rbldnsd_zones.c,v 1.61 2005/04/27 12:21:26 mjt Exp $
+/* $Id: rbldnsd_zones.c,v 1.69 2006/02/14 13:47:34 mjt Exp $
  * Nameserver zones: structures and routines
  */
 
@@ -9,7 +9,11 @@
 #include <syslog.h>
 #include <string.h>
 #include <errno.h>
+#include <unistd.h>
+#include <fcntl.h>
+#include <time.h>
 #include "rbldnsd.h"
+#include "istream.h"
 
 static struct dataset *ds_list;
 struct dataset *g_dsacl;
@@ -155,37 +159,43 @@ struct zone *addzone(struct zone *zonelist, const char *spec) {
   return zonelist;
 }
 
+/* helper routine for ds_special */
+static char *firstword(char *line, const char *word_lc) {
+  while(*word_lc)
+    if (dns_dnlc(*line) != *word_lc)
+      return NULL;
+    else
+      ++word_lc, ++line;
+  if (!ISSPACE(*line))
+    return NULL;
+  SKIPSPACE(line);
+  return line;
+}
+
 /* parse $SPECIAL construct */
-int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
-
-  switch(*line) {
-
-  case 's': case 'S':
-
-  if ((line[1] == 'o' || line[1] == 'O') &&
-      (line[2] == 'a' || line[2] == 'A') &&
-      ISSPACE(line[3]) && !isdstype(ds->ds_type, acl)) {
+static int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
+  char *w;
 
+  if ((w = firstword(line, "soa"))) {
     /* SOA record */
     struct dssoa dssoa;
     unsigned char odn[DNS_MAXDN], pdn[DNS_MAXDN];
     unsigned odnlen, pdnlen;
 
+    if (isdstype(ds->ds_type, acl))
+      return 0;	/* don't allow SOA for ACLs */
     if (ds->ds_dssoa)
       return 1; /* ignore if already set */
 
-    line += 4;
-    SKIPSPACE(line);
-
-    if (!(line = parse_ttl(line, &dssoa.dssoa_ttl, ds->ds_ttl))) return 0;
-    if (!(line = parse_dn(line, odn, &odnlen))) return 0;
-    if (!(line = parse_dn(line, pdn, &pdnlen))) return 0;
-    if (!(line = parse_uint32(line, &dssoa.dssoa_serial))) return 0;
-    if (!(line = parse_time_nb(line, dssoa.dssoa_n+0))) return 0;
-    if (!(line = parse_time_nb(line, dssoa.dssoa_n+4))) return 0;
-    if (!(line = parse_time_nb(line, dssoa.dssoa_n+8))) return 0;
-    if (!(line = parse_time_nb(line, dssoa.dssoa_n+12))) return 0;
-    if (*line) return 0;
+    if (!(w = parse_ttl(w, &dssoa.dssoa_ttl, ds->ds_ttl))) return 0;
+    if (!(w = parse_dn(w, odn, &odnlen))) return 0;
+    if (!(w = parse_dn(w, pdn, &pdnlen))) return 0;
+    if (!(w = parse_uint32(w, &dssoa.dssoa_serial))) return 0;
+    if (!(w = parse_time_nb(w, dssoa.dssoa_n+0))) return 0;
+    if (!(w = parse_time_nb(w, dssoa.dssoa_n+4))) return 0;
+    if (!(w = parse_time_nb(w, dssoa.dssoa_n+8))) return 0;
+    if (!(w = parse_time_nb(w, dssoa.dssoa_n+12))) return 0;
+    if (*w) return 0;
 
     dssoa.dssoa_odn = mp_memdup(ds->ds_mp, odn, odnlen);
     dssoa.dssoa_pdn = mp_memdup(ds->ds_mp, pdn, pdnlen);
@@ -193,16 +203,12 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
     ds->ds_dssoa = mp_talloc(ds->ds_mp, struct dssoa);
     if (!ds->ds_dssoa) return -1;
     *ds->ds_dssoa = dssoa;
-
     return 1;
   }
-  break;
-
-  case 'n': case 'N':
-
-  if ((line[1] == 's' || line[1] == 'S') &&
-      ISSPACE(line[2]) && !isdstype(ds->ds_type, acl)) {
 
+  if ((w = firstword(line, "ns")) ||
+      (w = firstword(line, "nameserver"))) {
+     /* NS records */
      unsigned char dn[DNS_MAXDN];
      unsigned dnlen;
      struct dsns *dsns, **dsnslp;
@@ -215,6 +221,9 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
      int newformat = 0;
 #endif
 
+    if (isdstype(ds->ds_type, acl))
+      return 0;	/* don't allow NSes for ACLs */
+
 #ifndef INCOMPAT_0_99
      if (ds->ds_nsflags & DSF_NEWNS) return 1;
      if (ds->ds_dsns) {
@@ -231,24 +240,21 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
      dsnslp = &ds->ds_dsns;
 #endif
 
-     line += 3;
-     SKIPSPACE(line);
-
      /*XXX parse options (AndrewSN suggested `-bloat') here */
 
-     if (!(line = parse_ttl(line, &ttl, ds->ds_ttl))) return 0;
+     if (!(w = parse_ttl(w, &ttl, ds->ds_ttl))) return 0;
 
      do {
-       if (*line == '-') {
+       if (*w == '-') {
          /* skip nameservers that start with `-' aka 'commented-out' */
-         do ++line; while (*line && !ISSPACE(*line));
-         SKIPSPACE(line);
+         do ++w; while (*w && !ISSPACE(*w));
+         SKIPSPACE(w);
 #ifndef INCOMPAT_0_99
 	 newformat = 1;
 #endif
          continue;
        }
-       if (!(line = parse_dn(line, dn, &dnlen))) return 0;
+       if (!(w = parse_dn(w, dn, &dnlen))) return 0;
        dsns = (struct dsns*)
          mp_alloc(ds->ds_mp, sizeof(struct dsns) + dnlen - 1, 1);
        if (!dsns) return -1;
@@ -260,7 +266,7 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
        if (!cnt++)
          dsns_first = dsns;
 #endif
-     } while(*line);
+     } while(*w);
 
 #ifndef INCOMPAT_0_99
      if (cnt > 1 || newformat) {
@@ -276,41 +282,24 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
 #else
      ds->ds_nsttl = ttl;
 #endif
-
-     return 1;
+    return 1;
   }
-  break;
 
-  case 't': case 'T':
-  if ((line[1] == 't' || line[1] == 'T') &&
-      (line[2] == 'l' || line[2] == 'L') &&
-      ISSPACE(line[3])) {
+  if ((w = firstword(line, "ttl"))) {
     unsigned ttl;
-    line += 4;
-    SKIPSPACE(line);
-    if (!(line = parse_ttl(line, &ttl, def_ttl))) return 0;
-    if (*line) return 0;
+    if (!(w = parse_ttl(w, &ttl, def_ttl))) return 0;
+    if (*w) return 0;
     if (dsc->dsc_subset) dsc->dsc_subset->ds_ttl = ttl;
     else ds->ds_ttl = ttl;
     return 1;
   }
-  break;
 
-  case 'm': case 'M':
-  if ((line[1] == 'A' || line[1] == 'a') &&
-      (line[2] == 'X' || line[2] == 'x') &&
-      (line[3] == 'R' || line[3] == 'r') &&
-      (line[4] == 'A' || line[4] == 'a') &&
-      (line[5] == 'N' || line[5] == 'n') &&
-      (line[6] == 'G' || line[6] == 'g') &&
-      (line[7] == 'E' || line[7] == 'e') &&
-      line[8] == '4' && ISSPACE(line[9])) {
+  if ((w = firstword(line, "maxrange4"))) {
     unsigned r;
     int cidr;
-    line += 10; SKIPSPACE(line);
-    if (*line == '/') cidr = 1, ++line;
+    if (*w == '/') cidr = 1, ++w;
     else cidr = 0;
-    if (!(line = parse_uint32(line, &r)) || *line || !r)
+    if (!(w = parse_uint32(w, &r)) || *w || !r)
       return 0;
     if (cidr) {
       if (r > 32) return 0;
@@ -324,41 +313,106 @@ int ds_special(struct dataset *ds, char *line, struct dsctx *dsc) {
     return 1;
   }
 
-  case '0': case '1': case '2': case '3': case '4':
-  case '5': case '6': case '7': case '8': case '9':
-  if (ISSPACE(line[1])) {
+  if (*(w = line) >= '0' && *w <= '9' && ISSPACE(w[1])) {
     /* substitution vars */
-    unsigned n = line[0] - '0';
+    unsigned n = w[0] - '0';
     if (dsc->dsc_subset) ds = dsc->dsc_subset;
     if (ds->ds_subst[n]) return 1; /* ignore second assignment */
-    line += 2;
-    SKIPSPACE(line);
-    if (!*line) return 0;
-    if (!(ds->ds_subst[n] = mp_strdup(ds->ds_mp, line))) return 0;
+    w += 2;
+    SKIPSPACE(w);
+    if (!*w) return 0;
+    if (!(ds->ds_subst[n] = mp_strdup(ds->ds_mp, w))) return 0;
     return 1;
   }
-  break;
 
-  case 'd': case 'D':
-  if ((line[1] == 'A' || line[1] == 'a') &&
-      (line[2] == 'T' || line[2] == 't') &&
-      (line[3] == 'A' || line[3] == 'a') &&
-      (line[4] == 'S' || line[4] == 's') &&
-      (line[5] == 'E' || line[5] == 'e') &&
-      (line[6] == 'T' || line[6] == 't') &&
-      ISSPACE(line[7]) &&
-      isdstype(ds->ds_type, combined)) {
-    line += 8;
-    SKIPSPACE(line);
-    return ds_combined_newset(ds, line, dsc);
+  if ((w = firstword(line, "dataset"))) {
+    if (!isdstype(ds->ds_type, combined))
+      return 0;	/* $dataset is only allowed for combined dataset */
+    return ds_combined_newset(ds, w, dsc);
   }
-  break;
 
+  if ((w = firstword(line, "timestamp"))) {
+    time_t stamp, expires;
+
+    if (!(w = parse_timestamp(w, &stamp))) return 0;
+    if (!*w)
+      expires = 0;
+    else if (*w == '+') {       /* relative */
+      unsigned n;
+      if (!(w = parse_time(w + 1, &n)) || *w) return 0;
+      if (!stamp || !n) return 0;
+      expires = stamp + n;
+      if (expires < 0 || expires - n != stamp) return 0;
+    }
+    else {
+      if (!(w = parse_timestamp(w, &expires)) || *w) return 0;
+    }
+    if (stamp) {
+      time_t now = time(NULL);
+      if (stamp > now) {
+        dslog(LOG_ERR, dsc,
+              "data timestamp is %u sec in the future, aborting loading",
+              (unsigned)(stamp - now));
+        return -1;
+      }
+    }
+    if (expires &&
+        (!ds->ds_expires || ds->ds_expires > expires))
+      ds->ds_expires = expires;
+    return 1;
   }
 
   return 0;
 }
 
+static int
+readdslines(struct istream *sp, struct dataset *ds, struct dsctx *dsc) {
+  char *line, *eol;
+  int r;
+  int noeol = 0;
+  struct dataset *dscur = ds;
+  ds_linefn_t *linefn = dscur->ds_type->dst_linefn;
+
+  while((r = istream_getline(sp, &line, '\n')) > 0) {
+    eol = line + r - 1;
+    if (noeol) {
+      if (*eol == '\n')
+        noeol = 0;
+      continue;
+    }
+    ++dsc->dsc_lineno;
+    if (*eol == '\n')
+      --eol;
+    else {
+      dswarn(dsc, "long line (truncated)");
+      noeol = 1; /* mark it to be read above */
+    }
+    SKIPSPACE(line);
+    while(eol >= line && ISSPACE(*eol))
+      --eol;
+    eol[1] = '\0';
+    if (line[0] == '$' ||
+        ((ISCOMMENT(line[0]) || line[0] == ':') && line[1] == '$')) {
+      int r = ds_special(ds, line[0] == '$' ? line + 1 : line + 2, dsc);
+      if (!r)
+        dswarn(dsc, "invalid or unrecognized special entry");
+      else if (r < 0)
+        return 0;
+      dscur = dsc->dsc_subset ? dsc->dsc_subset : ds;
+      linefn = dscur->ds_type->dst_linefn;
+      continue;
+    }
+    if (line[0] && !ISCOMMENT(line[0]))
+      if (!linefn(dscur, line, dsc))
+        return 0;
+  }
+  if (r < 0)
+    return -1;
+  if (noeol)
+    dslog(LOG_WARNING, dsc, "incomplete last line (ignored)");
+  return 1;
+}
+
 static void freedataset(struct dataset *ds) {
   ds->ds_type->dst_resetfn(ds->ds_dsd, 0);
   mp_free(ds->ds_mp);
@@ -372,10 +426,12 @@ static void freedataset(struct dataset *ds) {
   memset(ds->ds_subst, 0, sizeof(ds->ds_subst));
 }
 
-static int loaddataset(struct dataset *ds) {
+int loaddataset(struct dataset *ds) {
   struct dsfile *dsf;
   time_t stamp = 0;
-  FILE *f;
+  struct istream is;
+  int fd;
+  int r;
   struct stat st0, st1;
   struct dsctx dsc;
 
@@ -386,24 +442,43 @@ static int loaddataset(struct dataset *ds) {
 
   for(dsf = ds->ds_dsf; dsf; dsf = dsf->dsf_next) {
     dsc.dsc_fname = dsf->dsf_name;
-    f = fopen(dsf->dsf_name, "r");
-    if (!f || fstat(fileno(f), &st0) < 0) {
+    fd = open(dsf->dsf_name, O_RDONLY);
+    if (fd < 0 || fstat(fd, &st0) < 0) {
       dslog(LOG_ERR, &dsc, "unable to open file: %s", strerror(errno));
-      if (f) fclose(f);
-      return 0;
+      if (fd >= 0) close(fd);
+      goto fail;
     }
     ds->ds_type->dst_startfn(ds);
-    if (!readdslines(f, ds, &dsc)) {
-      fclose(f);
-      return 0;
+    istream_init_fd(&is, fd);
+    if (istream_compressed(&is)) {
+      if (nouncompress) {
+        dslog(LOG_ERR, &dsc, "file is compressed, decompression disabled");
+        r = 0;
+      }
+      else {
+#ifdef NO_ZLIB
+        dslog(LOG_ERR, &dsc,
+              "file is compressed, decompression is not compiled in");
+        r = 0;
+#else
+        r = istream_uncompress_setup(&is);
+          /* either 1 or -1 but not 0 */
+#endif
+      }
     }
+    else
+      r = 1;
+    if (r > 0) r = readdslines(&is, ds, &dsc);
+    if (r > 0) r = fstat(fd, &st1) < 0 ? -1 : 1;
     dsc.dsc_lineno = 0;
-    if (ferror(f) || fstat(fileno(f), &st1) < 0) {
+    istream_destroy(&is);
+    close(fd);
+    if (!r)
+      goto fail;
+    if (r < 0) {
       dslog(LOG_ERR, &dsc, "error reading file: %s", strerror(errno));
-      fclose(f);
-      return 0;
+      goto fail;
     }
-    fclose(f);
     if (st0.st_mtime != st1.st_mtime ||
         st0.st_size  != st1.st_size) {
       dslog(LOG_ERR, &dsc,
@@ -411,7 +486,7 @@ static int loaddataset(struct dataset *ds) {
       dslog(LOG_ERR, &dsc,
             "do not write data files directly, "
             "use temp file and rename(2) instead");
-      return 0;
+      goto fail;
     }
     dsf->dsf_stamp = st0.st_mtime;
     dsf->dsf_size  = st0.st_size;
@@ -424,97 +499,28 @@ static int loaddataset(struct dataset *ds) {
   ds->ds_type->dst_finishfn(ds, &dsc);
 
   return 1;
+
+fail:
+  freedataset(ds);
+  for (dsf = ds->ds_dsf; dsf; dsf = dsf->dsf_next)
+    dsf->dsf_stamp = 0;
+  ds->ds_stamp = 0;
+  return 0;
 }
 
-static int updatezone(struct zone *zone, const struct zone *zonelist) {
-  time_t stamp = 0;
-  const struct dssoa *dssoa = NULL;
-  const struct dsns *dsns = NULL;
-  unsigned nsttl = 0;
-  struct dslist *dsl;
-
-  for(dsl = zone->z_dsl; dsl; dsl = dsl->dsl_next) {
-    const struct dataset *ds = dsl->dsl_ds;
-    if (!ds->ds_stamp)
-      return 0;
-    if (stamp < ds->ds_stamp)
-      stamp = ds->ds_stamp;
-    if (!dssoa)
-      dssoa = ds->ds_dssoa;
-    if (!dsns)
-      dsns = ds->ds_dsns, nsttl = ds->ds_nsttl;
-  }
-
-  zone->z_stamp = stamp;
-  if (!update_zone_soa(zone, dssoa) ||
-      !update_zone_ns(zone, dsns, nsttl, zonelist))
-    zlog(LOG_WARNING, zone,
-         "NS or SOA RRs are too long, will be ignored");
-
-  return 1;
-}
-
-int reloadzones(struct zone *zonelist) {
-  struct dataset *ds;
+/* find next dataset which needs reloading */
+struct dataset *nextdataset2reload(struct dataset *ds) {
   struct dsfile *dsf;
-  struct zone *zone;
-  int reloaded = 0;
-  int errors = 0;
-  extern void start_loading();
-
-  for(ds = ds_list; ds; ds = ds->ds_next) {
-    int load = 0;
-
+  for (ds = ds ? ds->ds_next : ds_list; ds; ds = ds->ds_next)
     for(dsf = ds->ds_dsf; dsf; dsf = dsf->dsf_next) {
       struct stat st;
-      if (stat(dsf->dsf_name, &st) < 0) {
-        dslog(LOG_ERR, 0, "unable to stat file `%.60s': %s",
-              dsf->dsf_name, strerror(errno));
-        load = -1;
-        break;
-      }
-      else if (dsf->dsf_stamp != st.st_mtime ||
-               dsf->dsf_size  != st.st_size) {
-        load = 1;
-        dsf->dsf_stamp = st.st_mtime;
-        dsf->dsf_size = st.st_size;
-      }
+      if (stat(dsf->dsf_name, &st) < 0)
+        return ds;
+      if (dsf->dsf_stamp != st.st_mtime ||
+          dsf->dsf_size  != st.st_size)
+        return ds;
     }
-
-    if (!load)
-      continue;
-
-    ++reloaded;
-
-    if (load < 0 && !ds->ds_stamp) {
-      ++errors;
-      continue;
-    }
-
-    start_loading();
-
-    if (load < 0 || !loaddataset(ds)) {
-      ++errors;
-      freedataset(ds);
-      for (dsf = ds->ds_dsf; dsf; dsf = dsf->dsf_next)
-        dsf->dsf_stamp = 0;
-      ds->ds_stamp = 0;
-    }
-
-  }
-
-  if (reloaded) {
-
-    for(zone = zonelist; zone; zone = zone->z_next) {
-      if (!updatezone(zone, zonelist)) {
-        zlog(LOG_WARNING, zone, "zone will not be serviced");
-        zone->z_stamp = 0;
-      }
-    }
-
-  }
-
-  return errors ? -1 : reloaded ? 1 : 0;
+  return NULL;
 }
 
 #ifndef NO_MASTER_DUMP
